﻿using ProjectBoat.Drawnings;
using ProjectBoat.Exceptions;
using ProjectBoat.CollectionGenericObjects;
using ProjectBoat.Exceptions;
using System.Text;

namespace ProjectBoat.CollectionGenericObjects
{
    // Класс-хранилище коллекций
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class StorageCollection<T>
        where T : DrawningBoat
    {
        /// <summary>
        /// Словарь (хранилище) с коллекциями
        /// </summary>
        readonly Dictionary<CollectionInfo, ICollectionGenericObjects<T>> _storages;

        /// <summary>
        /// Возвращение списка названий коллекций
        /// </summary>
        public List<CollectionInfo> Keys => _storages.Keys.ToList();

        /// <summary>
        /// Ключевое слово, с которого должен начинаться файл
        /// </summary>
        private readonly string _collectionKey = "CollectionsStorage";

        /// <summary>
        /// Разделитель для записи ключа и значения элемента словаря
        /// </summary>
        private readonly string _separatorForKeyValue = "|";

        /// <summary>
        /// Разделитель для записей коллекции данных в файл
        /// </summary>
        private readonly string _separatorItems = ";";

        /// <summary>
        /// Конструктор
        /// </summary>
        public StorageCollection()
        {
            _storages = new Dictionary<CollectionInfo, ICollectionGenericObjects<T>>();
        }

        /// <summary>
        /// Добавление коллекции в хранилище
        /// </summary>
        /// <param name="name">Название коллекции</param>
        /// <param name="collectionType">тип коллекции</param>
        public void AddCollection(CollectionInfo name)
        {
            // TODO проверка, что name не пустой и нет в словаре записи с таким ключом
            // TODO Прописать логику для добавления

            if (name == null || _storages.ContainsKey(name))
            {
                return;
            }

            if (name.CollectionType == CollectionType.Massive)
            {
                _storages.Add(name, new MassiveGenericObjects<T>());
            }

            if (name.CollectionType == CollectionType.List)
            {
                _storages.Add(name, new ListGenericObjects<T>());
            }
        }

        /// <summary>
        /// Удаление коллекции
        /// </summary>
        /// <param name="name">Название коллекции</param>
        public void DelCollection(CollectionInfo name)
        {
            // TODO Прописать логику для удаления коллекции
            if (name == null || !_storages.ContainsKey(name))
            {
                return;
            }

            _storages.Remove(name);
        }

        /// <summary>
        /// Доступ к коллекции
        /// </summary>
        /// <param name="name">Название коллекции</param>
        /// <returns></returns>
        public ICollectionGenericObjects<T>? this[CollectionInfo name]
        {
            get
            {
                // TODO Продумать логику получения объекта
                if (_storages.ContainsKey(name))
                {
                    return _storages[name];
                }

                return null;
            }
        }

        /// <summary>
        /// Сохранение информации по самолетам в хранилище в файл
        /// </summary>
        /// <param name="filename">Путь и имя файла</param>
        /// <returns>true - сохранение прошло успешно, false - ошибка при сохранении данных</returns>
        public void SaveData(string filename)
        {
            if (_storages.Count == 0)
            {
                throw new Exception("В хранилище отсутствуют коллекции для сохранения");
            }


            if (File.Exists(filename))
            {
                File.Delete(filename);
            }

            using FileStream fs = new(filename, FileMode.Create);
            using StreamWriter streamWriter = new StreamWriter(fs);
            streamWriter.Write(_collectionKey);

            foreach (KeyValuePair<CollectionInfo, ICollectionGenericObjects<T>> value in _storages)
            {
                streamWriter.Write(Environment.NewLine);

                if (value.Value.Count == 0)
                {
                    continue;
                }

                streamWriter.Write(value.Key);
                streamWriter.Write(_separatorForKeyValue);
                streamWriter.Write(value.Value.MaxCount);
                streamWriter.Write(_separatorForKeyValue);


                foreach (T? item in value.Value.GetItems())
                {
                    string data = item?.GetDataForSave() ?? string.Empty;
                    if (string.IsNullOrEmpty(data))
                    {
                        continue;
                    }


                    streamWriter.Write(data);
                    streamWriter.Write(_separatorItems);

                }
            }
        }
        /// <summary>
        /// Загрузка информации по кораблям в хранилище из файла
        /// </summary>
        /// <param name="filename"></param>
        public void LoadData(string filename)
        {
            if (!File.Exists(filename))
            {
                throw new FileNotFoundException("Файл не существует");
            }

            using (StreamReader sr = new StreamReader(filename))
            {
                string? str;
                str = sr.ReadLine();
                if (str != _collectionKey.ToString())
                    throw new FormatException("В файле неверные данные");
                _storages.Clear();
                while ((str = sr.ReadLine()) != null)
                {
                    string[] record = str.Split(_separatorForKeyValue);
                    if (record.Length != 3)
                    {
                        continue;
                    }

                    CollectionInfo? collectionInfo = CollectionInfo.GetCollectionInfo(record[0]) ??
                    throw new Exception("Не удалось определить информацию коллекции: " + record[0]);
                    ICollectionGenericObjects<T>? collection = StorageCollection<T>.CreateCollection(collectionInfo.CollectionType) ??
                        throw new Exception("Не удалось создать коллекцию");
                    collection.MaxCount = Convert.ToInt32(record[1]);

                    string[] set = record[2].Split(_separatorItems, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string elem in set)
                    {
                        if (elem?.CreateDrawningBoat() is T boat)
                        {
                            try
                            {
                                if (collection.Insert(boat, new DrawiningBoatEqutables()) == -1)
                                {
                                    throw new InvalidOperationException("Объект не удалось добавить в коллекцию: " + record[3]);
                                }
                            }
                            catch (CollectionOverflowException ex)
                            {
                                throw new CollectionOverflowException("Коллекция переполнена", ex);
                            }
                            catch (ObjectAlreadyInCollectionException ex)
                            {
                                throw new InvalidOperationException("Объект уже присутствует в коллекции", ex);
                            }
                        }
                    }
                    _storages.Add(collectionInfo, collection);
                }
            }
        }

        /// <summary>
        /// Создание коллекции по типу
        /// </summary>
        /// <param name="collectionType"></param>
        /// <returns></returns>
        private static ICollectionGenericObjects<T>? CreateCollection(CollectionType collectionType)
        {
            return collectionType switch
            {
                CollectionType.Massive => new MassiveGenericObjects<T>(),
                CollectionType.List => new ListGenericObjects<T>(),
                _ => null,
            };
        }
    }
}
